{
  type lexem =
    | Atom_dot of string
    | Atom_turnstile of string
    | Atom_par of string
    | Atom_equal of string
    | Atom of string
    | Projector of string
    | Var of string
    | Int of int
    | Float of float
    | End_of_file
    | Par_open
    | Par_close
    | Bracket_open
    | Bracket_close
    | Double_bracket_open
    | Double_bracket_close
    | Curly_open
    | Curly_close
    | Minizinc of string
    | Include of string
    | String of string

  let print out_chan lexem =
    match lexem with
    | Atom_dot f -> Printf.fprintf out_chan "%s." f;
    | Atom_turnstile f -> Printf.fprintf out_chan "%s :-" f
    | Atom_equal f -> Printf.fprintf out_chan "%s =" f
    | Atom_par f -> Printf.fprintf out_chan "%s(" f
    | Atom f -> Printf.fprintf out_chan "%s" f
    | Projector f -> Printf.fprintf out_chan ".%s" f
    | Var x -> Printf.fprintf out_chan "%s" x
    | Int i -> Printf.fprintf out_chan "%d" i
    | Float f -> Printf.fprintf out_chan "%f" f
    | String s -> Printf.fprintf out_chan "%s" s
    | End_of_file -> Printf.fprintf out_chan "eof"
    | Par_open -> Printf.fprintf out_chan "("
    | Par_close -> Printf.fprintf out_chan ")"
    | Bracket_open -> Printf.fprintf out_chan "["
    | Bracket_close -> Printf.fprintf out_chan "]"
    | Double_bracket_open -> Printf.fprintf out_chan "[|"
    | Double_bracket_close -> Printf.fprintf out_chan "|]"
    | Curly_open -> Printf.fprintf out_chan "{"
    | Curly_close -> Printf.fprintf out_chan "}"
    | Minizinc s -> Printf.fprintf out_chan "minizinc %s" s
    | Include i -> Printf.fprintf out_chan "include %s" i

  type line_column = {
      line: int;
      column: int;
    }
    
  type position = {
      filename: string;
      start_p: line_column;
      end_p: line_column;
    }

  exception Error of position * string

  let position lexbuf =
    let start_p = Lexing.lexeme_start_p lexbuf in
    let end_p = Lexing.lexeme_end_p lexbuf in
    let get_line_column p = {
      line = p.Lexing.pos_lnum;
      column = p.Lexing.pos_cnum - p.Lexing.pos_bol
    } in {
    filename = start_p.Lexing.pos_fname;
    start_p = get_line_column start_p;
    end_p = get_line_column end_p
  }
    
  let minizinc_sz = 17

  let error position msg =
    raise (Error (position, msg))
}

let blank = [ ' ' '\t' ]

let lowercase = [ 'a' - 'z' ]

let uppercase = [ 'A' - 'Z' '_' ]

let digit = [ '0' - '9' ]

let ident_body = lowercase | uppercase | digit

let symbol = [ '.' '=' '+' '-' '*' '/' '\\' '<' '>' ':' '!' '#' ]
  
let name = lowercase ident_body*

let atom = name | symbol+

let var = uppercase ident_body*
    
rule main = parse
| blank {
    main lexbuf
  }
| "\n" {
    Lexing.new_line lexbuf;
    main lexbuf
  }
| ":-" {
    Atom ":-"
  }
| "include" {
    let s = string lexbuf in
    Include s
  }
| (atom as f) blank* "." {
    Atom_dot f
  }
| (atom as f) blank* ":-" {
    Atom_turnstile f
  }
| (atom as f) blank* "=" {
    Atom_equal f
  }
| (atom as f) "(" {
    Atom_par f
  }
| "var" {
    Atom "var"
  }
| "array" {
    Atom "array"
  }  
| "/*" {
    comment lexbuf;
    main lexbuf
  }
| "%" {
    comment_line lexbuf;
    main lexbuf
  }
| _ as c {
    let buffer = Buffer.create minizinc_sz in
    Buffer.add_char buffer c;
    minizinc buffer lexbuf
  }  
| eof {
    End_of_file
  }
and string = parse
| blank {
    string lexbuf
  }
| "\n" {
    Lexing.new_line lexbuf;
    string lexbuf
  }
| "\"" {
    let buffer = Buffer.create minizinc_sz in
    string_tail buffer lexbuf;
    Buffer.contents buffer
  }
and string_tail buffer = parse
| "\";" {
    ()
  }
| "\n" {
    Lexing.new_line lexbuf;
    string_tail buffer lexbuf
  }
| _ as c {
    Buffer.add_char buffer c;
    string_tail buffer lexbuf
  }
and minizinc buffer = parse
| ";" {
    Buffer.add_char buffer ';';
    Minizinc (Buffer.contents buffer)
  }
| eof {
    Minizinc (Buffer.contents buffer)
  }
| "\n" {
    Lexing.new_line lexbuf;
    Buffer.add_string buffer "\n";
    minizinc buffer lexbuf
  }
| "\"" {
    Buffer.add_string buffer "\"";
    minizinc_string buffer lexbuf;
    Buffer.add_string buffer "\"";
    minizinc buffer lexbuf
  }
| "/*" {
    comment lexbuf;
    minizinc buffer lexbuf
  }
| "%" {
    comment_line lexbuf;
    minizinc buffer lexbuf
  }
| _ as c {
    Buffer.add_char buffer c;
    minizinc buffer lexbuf
  }
and minizinc_string buffer = parse
| "\n" | eof {
    error (position lexbuf) "unterminated string"
  } 
| ("\\" _) as s {
    Buffer.add_string buffer s;
    minizinc_string buffer lexbuf
  }
| "\"" {
    ()
  } 
| _ as c {
    Buffer.add_char buffer c;
    minizinc_string buffer lexbuf
  } 
and term = parse
| blank {
    term lexbuf
  }
| "\n" {
    Lexing.new_line lexbuf;
    term lexbuf
  }
| "/*" {
    comment lexbuf;
    term lexbuf
  }
| "%" {
    comment_line lexbuf;
    term lexbuf
  }
| "." (name as f) {
    Projector f
  }
| atom as f {
    Atom f
  }
| (atom as f) "(" {
    Atom_par f
  }
| var as x {
    Var x
  }
| digit+ as s {
    Int (int_of_string s)
  }
| digit+ "." digit+ as s {
    Float (float_of_string s)
  }
| "," {
    Atom ","
  }
| ";" {
    Atom ";"
  }
| "|" {
    Atom "|"
  }
| "(" {
    Par_open
  }
| ")" {
    Par_close
  }
| "[" {
    Bracket_open
  }
| "]" {
    Bracket_close
  }
| "[|" {
    Double_bracket_open
  }
| "|]" {
    Double_bracket_close
  }
| "{" {
    Curly_open
  }
| "}" {
    Curly_close
  }
| "\"" {
    let buffer = Buffer.create minizinc_sz in
    minizinc_string buffer lexbuf;
    String (Buffer.contents buffer);
  }  
| eof {
    error (position lexbuf) "end of file before final dot"
  }
| _ as c {
    error (position lexbuf) (Printf.sprintf "unknown Prolog character %c" c) 
  }  
and comment = parse
| "*/" {
    ()
  }
| eof {
    error (position lexbuf) "end of file in comment"
  }
| "\n" {
    Lexing.new_line lexbuf;
    comment lexbuf
  }
| _ {
    comment lexbuf
  }
and comment_line = parse
| "\n" {
    Lexing.new_line lexbuf
  }
| eof {
    ()
  }
| _ {
    comment_line lexbuf
  }
